// Copyright (C) 2008-2012 Nikolaus Gebhardt
// This file is part of the "Irrlicht Engine".
// For conditions of distribution and use, see copyright notice in irrlicht.h

#pragma once

namespace video
{
enum E_INDEX_TYPE
{
	EIT_16BIT = 0,
	EIT_32BIT
};

} // end namespace video
