// Luanti
// SPDX-License-Identifier: LGPL-2.1-or-later
// Copyright (C) 2022 Minetest Authors

#pragma once

#include "config.h"

#if BUILD_BENCHMARKS
extern bool run_benchmarks(const char *arg = nullptr);
#endif
