#include <windows.h>
#include <winuser.h>
#include <commctrl.h>
#include <richedit.h>

#ifndef USE_CMAKE_CONFIG_H
#define USE_CMAKE_CONFIG_H
#endif
#include "config.h"
#undef USE_CMAKE_CONFIG_H

#if RUN_IN_PLACE
	#define BUILDMODE "RUN_IN_PLACE=1"
#else
	#define BUILDMODE "RUN_IN_PLACE=0"
#endif

#ifdef __MINGW32__
CREATEPROCESS_MANIFEST_RESOURCE_ID RT_MANIFEST "luanti.exe.manifest"
#endif

LANGUAGE 0, SUBLANG_NEUTRAL
130        ICON         "luanti-icon.ico"

/////////////////////////////////////////////////////////////////////////////
//
// Version
//

1 VERSIONINFO
 FILEVERSION VERSION_MAJOR,VERSION_MINOR,VERSION_PATCH,0
 PRODUCTVERSION VERSION_MAJOR,VERSION_MINOR,VERSION_PATCH,0
 FILEFLAGSMASK 0x3fL
#ifndef NDEBUG
 FILEFLAGS 0x1L
#else
 FILEFLAGS 0x0L
#endif
 FILEOS VOS_NT_WINDOWS32
 FILETYPE VFT_APP
 FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904b0"
    BEGIN
        VALUE "Comments", ""
        VALUE "CompanyName", PROJECT_NAME_C " community"
        VALUE "FileDescription", PROJECT_NAME_C " engine"
        VALUE "FileVersion", VERSION_STRING
        VALUE "InternalName", PROJECT_NAME
        VALUE "LegalCopyright", "(c) 2010-2024 Perttu Ahola (celeron55) and contributors"
        VALUE "OriginalFilename", PROJECT_NAME ".exe"
        VALUE "PrivateBuild", VERSION_EXTRA
        VALUE "ProductName", PROJECT_NAME_C
        VALUE "ProductVersion", PRODUCT_VERSION_STRING
        VALUE "SpecialBuild", BUILDMODE
    END
END
BLOCK "VarFileInfo"
BEGIN
    VALUE "Translation", 0x409, 1200
END
END

