// Luanti
// SPDX-License-Identifier: LGPL-2.1-or-later
// Copyright (C) 2013 celeron55, Perttu Ahola <celeron55@gmail.com>

#pragma once


/**
 * initialize basic default settings
 * @param settings pointer to settings
 */
void set_default_settings();
