// Luanti
// SPDX-License-Identifier: LGPL-2.1-or-later
// Copyright (C) 2013 celeron55, Perttu Ahola <celeron55@gmail.com>

#pragma once

extern const char *g_version_string;
extern const char *g_version_hash;
extern const char *g_build_info;
