// Luanti
// SPDX-License-Identifier: LGPL-2.1-or-later
// Copyright (C) 2025 grorp

#pragma once

#include "cpp_api/s_base.h"

class ScriptApiPauseMenu : virtual public ScriptApiBase
{
public:
	void open_settings();
};
