// Luanti
// SPDX-License-Identifier: LGPL-2.1-or-later
// Copyright (C) 2010-2013 celeron55, Perttu Ahola <celeron55@gmail.com>

#pragma once

#include "irrlichttypes.h"

#include "irr_v2d.h"
#include "irr_v3d.h"
#include "irr_aabb3d.h"

#include <SColor.h>
