// This file is automatically generated
// It contains a bunch of fake gettext calls, to tell xgettext about the strings in config files
// To update it, refer to the bottom of builtin/common/settings/init.lua

fake_function() {
	gettext("Controls");
	gettext("General");
	gettext("Camera smoothing");
	gettext("Smooths rotation of camera, also called look or mouse smoothing. 0 to disable.");
	gettext("Camera smoothing in cinematic mode");
	gettext("Smooths rotation of camera when in cinematic mode, 0 to disable. Enter cinematic mode by using the key set in Controls.");
	gettext("Build inside player");
	gettext("If enabled, you can place nodes at the position (feet + eye level) where you stand.\nThis is helpful when working with nodeboxes in small areas.");
	gettext("Aux1 key for climbing/descending");
	gettext("If enabled, \"Aux1\" key instead of \"Sneak\" key is used for climbing down and\ndescending.");
	gettext("Double tap jump for fly");
	gettext("Double-tapping the jump key toggles fly mode.");
	gettext("Always fly fast");
	gettext("If disabled, \"Aux1\" key is used to fly fast if both fly and fast mode are\nenabled.");
	gettext("Toggle Sneak key");
	gettext("If enabled, the \"Sneak\" key will toggle when pressed.\nThis functionality is ignored when fly is enabled.");
	gettext("Toggle Aux1 key");
	gettext("If enabled, the \"Aux1\" key will toggle when pressed.");
	gettext("Place repetition interval");
	gettext("The time in seconds it takes between repeated node placements when holding\nthe place button.");
	gettext("Minimum dig repetition interval");
	gettext("The minimum time in seconds it takes between digging nodes when holding\nthe dig button.");
	gettext("Automatic jumping");
	gettext("Automatically jump up single-node obstacles.");
	gettext("Safe digging and placing");
	gettext("Prevent digging and placing from repeating when holding the respective buttons.\nEnable this when you dig or place too often by accident.\nOn touchscreens, this only affects digging.");
	gettext("Keyboard and Mouse");
	gettext("Invert mouse");
	gettext("Invert vertical mouse movement.");
	gettext("Mouse sensitivity");
	gettext("Mouse sensitivity multiplier.");
	gettext("Hotbar: Enable mouse wheel for selection");
	gettext("Enable mouse wheel (scroll) for item selection in hotbar.");
	gettext("Hotbar: Invert mouse wheel direction");
	gettext("Invert mouse wheel (scroll) direction for item selection in hotbar.");
	gettext("Keybindings");
	gettext("Move forward");
	gettext("Move backward");
	gettext("Will also disable autoforward, when active.");
	gettext("Move left");
	gettext("Move right");
	gettext("Jump");
	gettext("Sneak");
	gettext("Also used for climbing down and descending in water if aux1_descends is disabled.");
	gettext("Dig/punch/use");
	gettext("Key for digging, punching or using something.\n(Note: The actual meaning might vary on a per-game basis.)");
	gettext("Place/use");
	gettext("Key for placing an item/block or for using something.\n(Note: The actual meaning might vary on a per-game basis.)");
	gettext("Open inventory");
	gettext("Aux1");
	gettext("Key for moving fast in fast mode.");
	gettext("Open chat");
	gettext("Command");
	gettext("Key for opening the chat window to type commands.");
	gettext("Local command");
	gettext("Key for opening the chat window to type local commands.");
	gettext("Toggle unlimited view range");
	gettext("Toggle fly");
	gettext("Toggle pitchmove");
	gettext("Toggle fast");
	gettext("Toggle noclip");
	gettext("Hotbar: select next item");
	gettext("Hotbar: select previous item");
	gettext("Mute");
	gettext("Increase volume");
	gettext("Decrease volume");
	gettext("Toggle automatic forward");
	gettext("Toggle cinematic mode");
	gettext("Toggle minimap");
	gettext("Screenshot");
	gettext("Toggle fullscreen");
	gettext("Drop item");
	gettext("Zoom");
	gettext("Toggle HUD");
	gettext("Toggle chat log");
	gettext("Toggle large chat console");
	gettext("Toggle fog");
	gettext("Toggle debug info");
	gettext("Toggle profiler");
	gettext("Key for toggling the display of the profiler. Used for development.");
	gettext("Toggle block bounds");
	gettext("Key for toggling the display of mapblock boundaries.");
	gettext("Toggle camera mode");
	gettext("Increase view range");
	gettext("Decrease view range");
	gettext("Return to Main Menu");
	gettext("Modifier key bind for closing your world.\nRequires ESC + the selected key to work.");
	gettext("Hotbar slot 1");
	gettext("Hotbar slot 2");
	gettext("Hotbar slot 3");
	gettext("Hotbar slot 4");
	gettext("Hotbar slot 5");
	gettext("Hotbar slot 6");
	gettext("Hotbar slot 7");
	gettext("Hotbar slot 8");
	gettext("Hotbar slot 9");
	gettext("Hotbar slot 10");
	gettext("Hotbar slot 11");
	gettext("Hotbar slot 12");
	gettext("Hotbar slot 13");
	gettext("Hotbar slot 14");
	gettext("Hotbar slot 15");
	gettext("Hotbar slot 16");
	gettext("Hotbar slot 17");
	gettext("Hotbar slot 18");
	gettext("Hotbar slot 19");
	gettext("Hotbar slot 20");
	gettext("Hotbar slot 21");
	gettext("Hotbar slot 22");
	gettext("Hotbar slot 23");
	gettext("Hotbar slot 24");
	gettext("Hotbar slot 25");
	gettext("Hotbar slot 26");
	gettext("Hotbar slot 27");
	gettext("Hotbar slot 28");
	gettext("Hotbar slot 29");
	gettext("Hotbar slot 30");
	gettext("Hotbar slot 31");
	gettext("Hotbar slot 32");
	gettext("Touchscreen");
	gettext("Touchscreen controls");
	gettext("Enables the touchscreen controls, allowing you to play the game with a touchscreen.\n\"auto\" means that the touchscreen controls will be enabled and disabled\nautomatically depending on the last used input method.");
	gettext("Interaction style");
	gettext("The kind of digging/placing controls used.\n\n* Tap\nLong/short tap anywhere on the screen to interact.\nInteraction happens at finger position.\n\n* Tap with crosshair\nLong/short tap anywhere on the screen to interact.\nInteraction happens at crosshair position.\n\n* Buttons with crosshair\nUse dedicated dig/place buttons to interact.\nInteraction happens at crosshair position.");
	gettext("Punch gesture");
	gettext("The gesture for punching players/entities.\nThis can be overridden by games and mods.\n\n* Short tap\nEasy to use and well-known from other games that shall not be named.\n\n* Long tap\nKnown from the classic Luanti mobile controls.\nCombat is more or less impossible.");
	gettext("Touchscreen sensitivity");
	gettext("Touchscreen sensitivity multiplier.");
	gettext("Movement threshold");
	gettext("The length in pixels after which a touch interaction is considered movement.");
	gettext("Threshold for long taps");
	gettext("The delay in milliseconds after which a touch interaction is considered a long tap.");
	gettext("Fixed virtual joystick");
	gettext("Fixes the position of virtual joystick.\nIf disabled, virtual joystick will center to first-touch's position.");
	gettext("Virtual joystick triggers Aux1 button");
	gettext("Use virtual joystick to trigger \"Aux1\" button.\nIf enabled, virtual joystick will also tap \"Aux1\" button when out of main circle.");
	gettext("Graphics and Audio");
	gettext("Graphics");
	gettext("Screen");
	gettext("Screen width");
	gettext("Width component of the initial window size.");
	gettext("Screen height");
	gettext("Height component of the initial window size.");
	gettext("Window maximized");
	gettext("Whether the window is maximized.");
	gettext("Remember screen size");
	gettext("Save window size automatically when modified.\nIf true, screen size is saved in screen_w and screen_h, and whether the window\nis maximized is stored in window_maximized.\n(Autosaving window_maximized only works if compiled with SDL.)");
	gettext("Full screen");
	gettext("Fullscreen mode.");
	gettext("Pause on lost window focus");
	gettext("Open the pause menu when the window's focus is lost. Does not pause if a formspec is\nopen.");
	gettext("FPS");
	gettext("Maximum FPS");
	gettext("If FPS would go higher than this, limit it by sleeping\nto not waste CPU power for no benefit.");
	gettext("VSync");
	gettext("Vertical screen synchronization. Your system may still force VSync on even if this is disabled.");
	gettext("FPS when unfocused");
	gettext("Maximum FPS when the window is not focused.");
	gettext("Viewing range");
	gettext("View distance in nodes.");
	gettext("Undersampling");
	gettext("Undersampling is similar to using a lower screen resolution, but it applies\nto the game world only, keeping the GUI intact.\nIt should give a significant performance boost at the cost of less detailed image.\nHigher values result in a less detailed image.\nNote: Undersampling is currently not supported if the \"3d_mode\" setting is set\nto a non-default value.");
	gettext("3D");
	gettext("3D mode");
	gettext("3D support.\nCurrently supported:\n-    none: no 3D output\n-    anaglyph: cyan/magenta color 3D\n-    topbottom: split screen top/bottom\n-    sidebyside: split screen side-by-side\n-    crossview: cross-eyed 3D");
	gettext("3D mode parallax strength");
	gettext("Strength of 3D mode parallax.");
	gettext("Bobbing");
	gettext("Arm inertia");
	gettext("Arm inertia, gives a more realistic movement of\nthe arm when the camera moves.");
	gettext("View bobbing factor");
	gettext("Enable view bobbing and amount of view bobbing.\nFor example: 0 for no view bobbing; 1.0 for normal; 2.0 for double.");
	gettext("Camera");
	gettext("Field of view");
	gettext("Field of view in degrees.");
	gettext("Light curve gamma");
	gettext("Alters the light curve by applying 'gamma correction' to it.\nHigher values make middle and lower light levels brighter.\nValue '1.0' leaves the light curve unaltered.\nThis only has significant effect on daylight and artificial\nlight, it has very little effect on natural night light.");
	gettext("Ambient occlusion gamma");
	gettext("The strength (darkness) of node ambient-occlusion shading.\nLower is darker, Higher is lighter. The valid range of values for this\nsetting is 0.25 to 4.0 inclusive. If the value is out of range it will be\nset to the nearest valid value.");
	gettext("Screenshots");
	gettext("Screenshot folder");
	gettext("Path to save screenshots at. Can be an absolute or relative path.\nThe folder will be created if it doesn't already exist.");
	gettext("Screenshot format");
	gettext("Format of screenshots.");
	gettext("Screenshot quality");
	gettext("Screenshot quality. Only used for JPEG format.\n1 means worst quality; 100 means best quality.\nUse 0 for default quality.");
	gettext("Node and Entity Highlighting");
	gettext("Node highlighting");
	gettext("Method used to highlight selected object.");
	gettext("Show entity selection boxes");
	gettext("Show entity selection boxes\nA restart is required after changing this.");
	gettext("Selection box color");
	gettext("Selection box border color (R,G,B).");
	gettext("Selection box width");
	gettext("Width of the selection box lines around nodes.");
	gettext("Crosshair color");
	gettext("Crosshair color (R,G,B).\nAlso controls the object crosshair color");
	gettext("Crosshair alpha");
	gettext("Crosshair alpha (opaqueness, between 0 and 255).\nThis also applies to the object crosshair.");
	gettext("Fog");
	gettext("Whether to fog out the end of the visible area.");
	gettext("Colored fog");
	gettext("Make fog and sky colors depend on daytime (dawn/sunset) and view direction.");
	gettext("Fog start");
	gettext("Fraction of the visible distance at which fog starts to be rendered");
	gettext("Clouds");
	gettext("3D clouds");
	gettext("Allow clouds to look 3D instead of flat.");
	gettext("Soft clouds");
	gettext("Use smooth cloud shading.");
	gettext("Filtering and Antialiasing");
	gettext("Mipmapping");
	gettext("Use mipmaps when scaling textures. May slightly increase performance,\nespecially when using a high-resolution texture pack.\nGamma-correct downscaling is not supported.");
	gettext("Bilinear filtering");
	gettext("Use bilinear filtering when scaling textures.");
	gettext("Trilinear filtering");
	gettext("Use trilinear filtering when scaling textures.\nIf both bilinear and trilinear filtering are enabled, trilinear filtering\nis applied.");
	gettext("Anisotropic filtering");
	gettext("Use anisotropic filtering when looking at textures from an angle.\nThis provides a significant improvement when used together with mipmapping.");
	gettext("Antialiasing method");
	gettext("Select the antialiasing method to apply.\n\n* None - No antialiasing (default)\n\n* FSAA - Hardware-provided full-screen antialiasing\nA.K.A multi-sample antialiasing (MSAA)\nSmoothens out block edges but does not affect the insides of textures.\n\nIf Post Processing is disabled, changing FSAA requires a restart.\nAlso, if Post Processing is disabled, FSAA will not work together with\nundersampling or a non-default \"3d_mode\" setting.\n\n* FXAA - Fast approximate antialiasing\nApplies a post-processing filter to detect and smoothen high-contrast edges.\nProvides balance between speed and image quality.\n\n* SSAA - Super-sampling antialiasing\nRenders higher-resolution image of the scene, then scales down to reduce\nthe aliasing effects. This is the slowest and the most accurate method.");
	gettext("Anti-aliasing scale");
	gettext("Defines the size of the sampling grid for FSAA and SSAA antialiasing methods.\nValue of 2 means taking 2x2 = 4 samples.");
	gettext("Occlusion Culling");
	gettext("Occlusion Culler");
	gettext("Type of occlusion_culler\n\n\"loops\" is the legacy algorithm with nested loops and O(n³) complexity\n\"bfs\" is the new algorithm based on breadth-first-search and side culling\n\nThis setting should only be changed if you have performance problems.");
	gettext("Enable Raytraced Culling");
	gettext("Use raytraced occlusion culling in the new culler.\nThis flag enables use of raytraced occlusion culling test for\nclient mesh sizes smaller than 4x4x4 map blocks.");
	gettext("Effects");
	gettext("Translucent liquids");
	gettext("Allows liquids to be translucent.");
	gettext("Leaves style");
	gettext("Leaves style:\n-   Fancy:  all faces visible\n-   Simple: only outer faces\n-   Opaque: disable transparency");
	gettext("Connect glass");
	gettext("Connects glass if supported by node.");
	gettext("Smooth lighting");
	gettext("Enable smooth lighting with simple ambient occlusion.");
	gettext("Tradeoffs for performance");
	gettext("Enables tradeoffs that reduce CPU load or increase rendering performance\nat the expense of minor visual glitches that do not impact game playability.");
	gettext("Waving Nodes");
	gettext("Waving leaves");
	gettext("Set to true to enable waving leaves.");
	gettext("Waving plants");
	gettext("Set to true to enable waving plants.");
	gettext("Waving liquids");
	gettext("Set to true to enable waving liquids (like water).");
	gettext("Waving liquids wave height");
	gettext("The maximum height of the surface of waving liquids.\n4.0 = Wave height is two nodes.\n0.0 = Wave doesn't move at all.\nDefault is 1.0 (1/2 node).");
	gettext("Waving liquids wavelength");
	gettext("Length of liquid waves.");
	gettext("Waving liquids wave speed");
	gettext("How fast liquid waves will move. Higher = faster.\nIf negative, liquid waves will move backwards.");
	gettext("Dynamic shadows");
	gettext("Set to true to enable Shadow Mapping.");
	gettext("Shadow strength gamma");
	gettext("Set the shadow strength gamma.\nAdjusts the intensity of in-game dynamic shadows.\nLower value means lighter shadows, higher value means darker shadows.");
	gettext("Shadow map max distance in nodes to render shadows");
	gettext("Maximum distance to render shadows.");
	gettext("Shadow map texture size");
	gettext("Texture size to render the shadow map on.\nThis must be a power of two.\nBigger numbers create better shadows but it is also more expensive.");
	gettext("Shadow map texture in 32 bits");
	gettext("Sets shadow texture quality to 32 bits.\nOn false, 16 bits texture will be used.\nThis can cause much more artifacts in the shadow.");
	gettext("Shadow filter quality");
	gettext("Define shadow filtering quality.\nThis simulates the soft shadows effect by applying a PCF or Poisson disk\nbut also uses more resources.");
	gettext("Colored shadows");
	gettext("Enable colored shadows for transculent nodes.\nThis is expensive.");
	gettext("Soft shadow radius");
	gettext("Set the soft shadow radius size.\nLower values mean sharper shadows, bigger values mean softer shadows.\nMinimum value: 1.0; maximum value: 15.0");
	gettext("Sky Body Orbit Tilt");
	gettext("Set the default tilt of Sun/Moon orbit in degrees.\nGames may change orbit tilt via API.\nValue of 0 means no tilt / vertical orbit.");
	gettext("Post Processing");
	gettext("Enable Post Processing");
	gettext("Enables the post processing pipeline.");
	gettext("Filmic tone mapping");
	gettext("Enables Hable's 'Uncharted 2' filmic tone mapping.\nSimulates the tone curve of photographic film and how this approximates the\nappearance of high dynamic range images. Mid-range contrast is slightly\nenhanced, highlights and shadows are gradually compressed.");
	gettext("Enable Automatic Exposure");
	gettext("Enable automatic exposure correction\nWhen enabled, the post-processing engine will\nautomatically adjust to the brightness of the scene,\nsimulating the behavior of human eye.");
	gettext("Exposure compensation");
	gettext("Set the exposure compensation in EV units.\nValue of 0.0 (default) means no exposure compensation.\nRange: from -1 to 1.0");
	gettext("Enable Debanding");
	gettext("Apply dithering to reduce color banding artifacts.\nDithering significantly increases the size of losslessly-compressed\nscreenshots and it works incorrectly if the display or operating system\nperforms additional dithering or if the color channels are not quantized\nto 8 bits.\nWith OpenGL ES, dithering only works if the shader supports high\nfloating-point precision and it may have a higher performance impact.");
	gettext("Enable Bloom");
	gettext("Set to true to enable bloom effect.\nBright colors will bleed over the neighboring objects.");
	gettext("Volumetric lighting");
	gettext("Set to true to enable volumetric lighting effect (a.k.a. \"Godrays\").");
	gettext("Other Effects");
	gettext("Translucent foliage");
	gettext("Simulate translucency when looking at foliage in the sunlight.");
	gettext("Liquid reflections");
	gettext("When enabled, liquid reflections are simulated.");
	gettext("Audio");
	gettext("Volume");
	gettext("Volume of all sounds.\nRequires the sound system to be enabled.");
	gettext("Volume when unfocused");
	gettext("Volume multiplier when the window is unfocused.");
	gettext("Mute sound");
	gettext("Whether to mute sounds. You can unmute sounds at any time.\nIn-game, you can toggle the mute state with the mute key or by using the\npause menu.");
	gettext("User Interfaces");
	gettext("Language");
	gettext("Set the language. By default, the system language is used.\nA restart is required after changing this.");
	gettext("GUI");
	gettext("Optimize GUI for touchscreens");
	gettext("When enabled, the GUI is optimized to be more usable on touchscreens.\nWhether this is enabled by default depends on your hardware form-factor.");
	gettext("GUI scaling");
	gettext("Scale GUI by a user specified value.\nUse a nearest-neighbor-anti-alias filter to scale the GUI.\nThis will smooth over some of the rough edges, and blend\npixels when scaling down, at the cost of blurring some\nedge pixels when images are scaled by non-integer sizes.");
	gettext("Smooth scrolling");
	gettext("Enables smooth scrolling.");
	gettext("Inventory items animations");
	gettext("Enables animation of inventory items.");
	gettext("Formspec Full-Screen Background Opacity");
	gettext("Formspec full-screen background opacity (between 0 and 255).");
	gettext("Formspec Full-Screen Background Color");
	gettext("Formspec full-screen background color (R,G,B).");
	gettext("GUI scaling filter");
	gettext("When gui_scaling_filter is true, all GUI images need to be\nfiltered in software, but some images are generated directly\nto hardware (e.g. render-to-texture for nodes in inventory).");
	gettext("Tooltip delay");
	gettext("Delay showing tooltips, stated in milliseconds.");
	gettext("Append item name");
	gettext("Append item name to tooltip.");
	gettext("Clouds in menu");
	gettext("Use a cloud animation for the main menu background.");
	gettext("HUD");
	gettext("HUD scaling");
	gettext("Modifies the size of the HUD elements.");
	gettext("Show name tag backgrounds by default");
	gettext("Whether name tag backgrounds should be shown by default.\nMods may still set a background.");
	gettext("Show debug info");
	gettext("Whether to show the client debug info (has the same effect as hitting F5).");
	gettext("Block bounds HUD radius");
	gettext("Radius to use when the block bounds HUD feature is set to near blocks.");
	gettext("Maximum hotbar width");
	gettext("Maximum proportion of current window to be used for hotbar.\nUseful if there's something to be displayed right or left of hotbar.");
	gettext("Chat");
	gettext("Recent Chat Messages");
	gettext("Maximum number of recent chat messages to show");
	gettext("Console height");
	/* xgettext:no-c-format */ gettext("In-game chat console height, between 0.1 (10%) and 1.0 (100%).");
	gettext("Console color");
	gettext("In-game chat console background color (R,G,B).");
	gettext("Console alpha");
	gettext("In-game chat console background alpha (opaqueness, between 0 and 255).");
	gettext("Weblink color");
	gettext("Optional override for chat weblink color.");
	gettext("Chat font size");
	gettext("Font size of the recent chat text and chat prompt in point (pt).\nValue 0 will use the default font size.");
	gettext("Content Repository");
	gettext("ContentDB URL");
	gettext("The URL for the content repository");
	gettext("Enable updates available indicator on content tab");
	gettext("If enabled and you have ContentDB packages installed, Luanti may contact ContentDB to\ncheck for package updates when opening the mainmenu.");
	gettext("ContentDB Flag Blacklist");
	gettext("Comma-separated list of flags to hide in the content repository.\n\"nonfree\" can be used to hide packages which do not qualify as 'free software',\nas defined by the Free Software Foundation.\nYou can also specify content ratings.\nThese flags are independent from Luanti versions,\nso see a full list at https://content.luanti.org/help/content_flags/");
	gettext("ContentDB Max Concurrent Downloads");
	gettext("Maximum number of concurrent downloads. Downloads exceeding this limit will be queued.\nThis should be lower than curl_parallel_limit.");
	gettext("Client and Server");
	gettext("Client");
	gettext("Saving map received from server");
	gettext("Save the map received by the client on disk.");
	gettext("Serverlist URL");
	gettext("URL to the server list displayed in the Multiplayer Tab.");
	gettext("Enable split login/register");
	gettext("If enabled, server account registration is separate from login in the UI.\nIf disabled, connecting to a server will automatically register a new account.");
	gettext("Update information URL");
	gettext("URL to JSON file which provides information about the newest Luanti release.\nIf this is empty the engine will never check for updates.");
	gettext("Server");
	gettext("Admin name");
	gettext("Name of the player.\nWhen running a server, a client connecting with this name is admin.\nWhen starting from the main menu, this is overridden.");
	gettext("Serverlist and MOTD");
	gettext("Server name");
	gettext("Name of the server, to be displayed when players join and in the serverlist.");
	gettext("Server description");
	gettext("Description of server, to be displayed when players join and in the serverlist.");
	gettext("Server address");
	gettext("Domain name of server, to be displayed in the serverlist.");
	gettext("Server URL");
	gettext("Homepage of server, to be displayed in the serverlist.");
	gettext("Announce server");
	gettext("Automatically report to the serverlist.");
	gettext("Send player names to the server list");
	gettext("Send names of online players to the serverlist. If disabled only the player count is revealed.");
	gettext("Announce to this serverlist.");
	gettext("Message of the day");
	gettext("Message of the day displayed to players connecting.");
	gettext("Maximum users");
	gettext("Maximum number of players that can be connected simultaneously.");
	gettext("Static spawn point");
	gettext("If this is set, players will always (re)spawn at the given position.");
	gettext("Networking");
	gettext("Server port");
	gettext("Network port to listen (UDP).\nThis value will be overridden when starting from the main menu.");
	gettext("Bind address");
	gettext("The network interface that the server listens on.");
	gettext("Strict protocol checking");
	gettext("Enable to disallow old clients from connecting.\nOlder clients are compatible in the sense that they will not crash when connecting\nto new servers, but they may not support all new features that you are expecting.");
	gettext("Protocol version minimum");
	gettext("Define the oldest clients allowed to connect.\nOlder clients are compatible in the sense that they will not crash when connecting\nto new servers, but they may not support all new features that you are expecting.\nThis allows for more fine-grained control than strict_protocol_version_checking.\nLuanti still enforces its own internal minimum, and enabling\nstrict_protocol_version_checking will effectively override this.");
	gettext("Remote media");
	gettext("Specifies URL from which client fetches media instead of using UDP.\n$filename should be accessible from $remote_media$filename via cURL\n(obviously, remote_media should end with a slash).\nFiles that are not present will be fetched the usual way.");
	gettext("IPv6 server");
	gettext("Enable IPv6 support for server.\nNote that clients will be able to connect with both IPv4 and IPv6.\nIgnored if bind_address is set.");
	gettext("Server Security");
	gettext("Default password");
	gettext("New users need to input this password.");
	gettext("Disallow empty passwords");
	gettext("If enabled, players cannot join without a password or change theirs to an empty password.");
	gettext("Default privileges");
	gettext("The privileges that new users automatically get.\nSee /privs in game for a full list on your server and mod configuration.");
	gettext("Basic privileges");
	gettext("Privileges that players with basic_privs can grant");
	gettext("Anticheat flags");
	gettext("Server anticheat configuration.\nFlags are positive. Uncheck the flag to disable corresponding anticheat module.");
	gettext("Anticheat movement tolerance");
	gettext("Tolerance of movement cheat detector.\nIncrease the value if players experience stuttery movement.");
	gettext("Rollback recording");
	gettext("If enabled, actions are recorded for rollback.\nThis option is only read when server starts.");
	gettext("Client-side Modding");
	gettext("Client side modding restrictions");
	gettext("Restricts the access of certain client-side functions on servers.\nCombine the byteflags below to restrict client-side features, or set to 0\nfor no restrictions:\nLOAD_CLIENT_MODS: 1 (disable loading client-provided mods)\nCHAT_MESSAGES: 2 (disable send_chat_message call client-side)\nREAD_ITEMDEFS: 4 (disable get_item_def call client-side)\nREAD_NODEDEFS: 8 (disable get_node_def call client-side)\nLOOKUP_NODES_LIMIT: 16 (limits get_node call client-side to\ncsm_restriction_noderange)\nREAD_PLAYERINFO: 32 (disable get_player_names call client-side)");
	gettext("Client-side node lookup range restriction");
	gettext("If the CSM restriction for node range is enabled, get_node calls are limited\nto this distance from the player to the node.");
	gettext("Strip color codes");
	gettext("Remove color codes from incoming chat messages\nUse this to stop players from being able to use color in their messages");
	gettext("Chat message max length");
	gettext("Set the maximum length of a chat message (in characters) sent by clients.");
	gettext("Chat message count limit");
	gettext("Number of messages a player may send per 10 seconds.");
	gettext("Chat message kick threshold");
	gettext("Kick players who sent more than X messages per 10 seconds.");
	gettext("Server Gameplay");
	gettext("Time speed");
	gettext("Controls length of day/night cycle.\nExamples:\n72 = 20min, 360 = 4min, 1 = 24hour, 0 = day/night/whatever stays unchanged.");
	gettext("World start time");
	gettext("Time of day when a new world is started, in millihours (0-23999).");
	gettext("Item entity TTL");
	gettext("Time in seconds for item entity (dropped items) to live.\nSetting it to -1 disables the feature.");
	gettext("Default stack size");
	gettext("Specifies the default stack size of nodes, items and tools.\nNote that mods or games may explicitly set a stack for certain (or all) items.");
	gettext("Physics");
	gettext("Default acceleration");
	gettext("Horizontal and vertical acceleration on ground or when climbing,\nin nodes per second per second.");
	gettext("Acceleration in air");
	gettext("Horizontal acceleration in air when jumping or falling,\nin nodes per second per second.");
	gettext("Fast mode acceleration");
	gettext("Horizontal and vertical acceleration in fast mode,\nin nodes per second per second.");
	gettext("Walking speed");
	gettext("Walking and flying speed, in nodes per second.");
	gettext("Sneaking speed");
	gettext("Sneaking speed, in nodes per second.");
	gettext("Fast mode speed");
	gettext("Walking, flying and climbing speed in fast mode, in nodes per second.");
	gettext("Climbing speed");
	gettext("Vertical climbing speed, in nodes per second.");
	gettext("Jumping speed");
	gettext("Initial vertical speed when jumping, in nodes per second.");
	gettext("Liquid fluidity");
	gettext("How much you are slowed down when moving inside a liquid.\nDecrease this to increase liquid resistance to movement.");
	gettext("Liquid fluidity smoothing");
	gettext("Maximum liquid resistance. Controls deceleration when entering liquid at\nhigh speed.");
	gettext("Liquid sinking");
	gettext("Controls sinking speed in liquid when idling. Negative values will cause\nyou to rise instead.");
	gettext("Gravity");
	gettext("Acceleration of gravity, in nodes per second per second.");
	gettext("Mapgen");
	gettext("Fixed map seed");
	gettext("A chosen map seed for a new map, leave empty for random.\nWill be overridden when creating a new world in the main menu.");
	gettext("Mapgen name");
	gettext("Name of map generator to be used when creating a new world.\nCreating a world in the main menu will override this.\nCurrent mapgens in a highly unstable state:\n-    The optional floatlands of v7 (disabled by default).");
	gettext("Water level");
	gettext("Water surface level of the world.");
	gettext("Max block generate distance");
	gettext("From how far blocks are generated for clients, stated in mapblocks (16 nodes).");
	gettext("Map generation limit");
	gettext("Limit of map generation, in nodes, in all 6 directions from (0, 0, 0).\nOnly mapchunks completely within the mapgen limit are generated.\nValue is stored per-world.");
	gettext("Mapgen flags");
	gettext("Global map generation attributes.\nIn Mapgen v6 the 'decorations' flag controls all decorations except trees\nand jungle grass, in all other mapgens this flag controls all decorations.");
	gettext("Biome API");
	gettext("Heat noise");
	gettext("Temperature variation for biomes.");
	gettext("Heat blend noise");
	gettext("Small-scale temperature variation for blending biomes on borders.");
	gettext("Humidity noise");
	gettext("Humidity variation for biomes.");
	gettext("Humidity blend noise");
	gettext("Small-scale humidity variation for blending biomes on borders.");
	gettext("Mapgen V5");
	gettext("Mapgen V5 specific flags");
	gettext("Map generation attributes specific to Mapgen v5.");
	gettext("Cave width");
	gettext("Controls width of tunnels, a smaller value creates wider tunnels.\nValue >= 10.0 completely disables generation of tunnels and avoids the\nintensive noise calculations.");
	gettext("Large cave depth");
	gettext("Y of upper limit of large caves.");
	gettext("Small cave minimum number");
	gettext("Minimum limit of random number of small caves per mapchunk.");
	gettext("Small cave maximum number");
	gettext("Maximum limit of random number of small caves per mapchunk.");
	gettext("Large cave minimum number");
	gettext("Minimum limit of random number of large caves per mapchunk.");
	gettext("Large cave maximum number");
	gettext("Maximum limit of random number of large caves per mapchunk.");
	gettext("Large cave proportion flooded");
	gettext("Proportion of large caves that contain liquid.");
	gettext("Cavern limit");
	gettext("Y-level of cavern upper limit.");
	gettext("Cavern taper");
	gettext("Y-distance over which caverns expand to full size.");
	gettext("Cavern threshold");
	gettext("Defines full size of caverns, smaller values create larger caverns.");
	gettext("Dungeon minimum Y");
	gettext("Lower Y limit of dungeons.");
	gettext("Dungeon maximum Y");
	gettext("Upper Y limit of dungeons.");
	gettext("Noises");
	gettext("Filler depth noise");
	gettext("Variation of biome filler depth.");
	gettext("Factor noise");
	gettext("Variation of terrain vertical scale.\nWhen noise is < -0.55 terrain is near-flat.");
	gettext("Height noise");
	gettext("Y-level of average terrain surface.");
	gettext("Cave1 noise");
	gettext("First of two 3D noises that together define tunnels.");
	gettext("Cave2 noise");
	gettext("Second of two 3D noises that together define tunnels.");
	gettext("Cavern noise");
	gettext("3D noise defining giant caverns.");
	gettext("Ground noise");
	gettext("3D noise defining terrain.");
	gettext("Dungeon noise");
	gettext("3D noise that determines number of dungeons per mapchunk.");
	gettext("Mapgen V6");
	gettext("Mapgen V6 specific flags");
	gettext("Map generation attributes specific to Mapgen v6.\nThe 'snowbiomes' flag enables the new 5 biome system.\nWhen the 'snowbiomes' flag is enabled jungles are automatically enabled and\nthe 'jungles' flag is ignored.\nThe 'temples' flag disables generation of desert temples. Normal dungeons will appear instead.");
	gettext("Desert noise threshold");
	gettext("Deserts occur when np_biome exceeds this value.\nWhen the 'snowbiomes' flag is enabled, this is ignored.");
	gettext("Beach noise threshold");
	gettext("Sandy beaches occur when np_beach exceeds this value.");
	gettext("Terrain base noise");
	gettext("Y-level of lower terrain and seabed.");
	gettext("Terrain higher noise");
	gettext("Y-level of higher terrain that creates cliffs.");
	gettext("Steepness noise");
	gettext("Varies steepness of cliffs.");
	gettext("Height select noise");
	gettext("Defines distribution of higher terrain.");
	gettext("Mud noise");
	gettext("Varies depth of biome surface nodes.");
	gettext("Beach noise");
	gettext("Defines areas with sandy beaches.");
	gettext("Biome noise");
	gettext("Cave noise");
	gettext("Variation of number of caves.");
	gettext("Trees noise");
	gettext("Defines tree areas and tree density.");
	gettext("Apple trees noise");
	gettext("Defines areas where trees have apples.");
	gettext("Mapgen V7");
	gettext("Mapgen V7 specific flags");
	gettext("Map generation attributes specific to Mapgen v7.\n'ridges': Rivers.\n'floatlands': Floating land masses in the atmosphere.\n'caverns': Giant caves deep underground.");
	gettext("Mountain zero level");
	gettext("Y of mountain density gradient zero level. Used to shift mountains vertically.");
	gettext("Floatland minimum Y");
	gettext("Lower Y limit of floatlands.");
	gettext("Floatland maximum Y");
	gettext("Upper Y limit of floatlands.");
	gettext("Floatland tapering distance");
	gettext("Y-distance over which floatlands taper from full density to nothing.\nTapering starts at this distance from the Y limit.\nFor a solid floatland layer, this controls the height of hills/mountains.\nMust be less than or equal to half the distance between the Y limits.");
	gettext("Floatland taper exponent");
	gettext("Exponent of the floatland tapering. Alters the tapering behavior.\nValue = 1.0 creates a uniform, linear tapering.\nValues > 1.0 create a smooth tapering suitable for the default separated\nfloatlands.\nValues < 1.0 (for example 0.25) create a more defined surface level with\nflatter lowlands, suitable for a solid floatland layer.");
	gettext("Floatland density");
	/* xgettext:no-c-format */ gettext("Adjusts the density of the floatland layer.\nIncrease value to increase density. Can be positive or negative.\nValue = 0.0: 50% of volume is floatland.\nValue = 2.0 (can be higher depending on 'mgv7_np_floatland', always test\nto be sure) creates a solid floatland layer.");
	gettext("Floatland water level");
	gettext("Surface level of optional water placed on a solid floatland layer.\nWater is disabled by default and will only be placed if this value is set\nto above 'mgv7_floatland_ymax' - 'mgv7_floatland_taper' (the start of the\nupper tapering).\n***WARNING, POTENTIAL DANGER TO WORLDS AND SERVER PERFORMANCE***:\nWhen enabling water placement, floatlands must be configured and tested\nto be a solid layer by setting 'mgv7_floatland_density' to 2.0 (or other\nrequired value depending on 'mgv7_np_floatland'), to avoid\nserver-intensive extreme water flow and to avoid vast flooding of the\nworld surface below.");
	gettext("Terrain alternative noise");
	gettext("Terrain persistence noise");
	gettext("Varies roughness of terrain.\nDefines the 'persistence' value for terrain_base and terrain_alt noises.");
	gettext("Defines distribution of higher terrain and steepness of cliffs.");
	gettext("Mountain height noise");
	gettext("Variation of maximum mountain height (in nodes).");
	gettext("Ridge underwater noise");
	gettext("Defines large-scale river channel structure.");
	gettext("Mountain noise");
	gettext("3D noise defining mountain structure and height.\nAlso defines structure of floatland mountain terrain.");
	gettext("Ridge noise");
	gettext("3D noise defining structure of river canyon walls.");
	gettext("Floatland noise");
	gettext("3D noise defining structure of floatlands.\nIf altered from the default, the noise 'scale' (0.7 by default) may need\nto be adjusted, as floatland tapering functions best when this noise has\na value range of approximately -2.0 to 2.0.");
	gettext("Mapgen Carpathian");
	gettext("Mapgen Carpathian specific flags");
	gettext("Map generation attributes specific to Mapgen Carpathian.");
	gettext("Base ground level");
	gettext("Defines the base ground level.");
	gettext("River channel width");
	gettext("Defines the width of the river channel.");
	gettext("River channel depth");
	gettext("Defines the depth of the river channel.");
	gettext("River valley width");
	gettext("Defines the width of the river valley.");
	gettext("Hilliness1 noise");
	gettext("First of 4 2D noises that together define hill/mountain range height.");
	gettext("Hilliness2 noise");
	gettext("Second of 4 2D noises that together define hill/mountain range height.");
	gettext("Hilliness3 noise");
	gettext("Third of 4 2D noises that together define hill/mountain range height.");
	gettext("Hilliness4 noise");
	gettext("Fourth of 4 2D noises that together define hill/mountain range height.");
	gettext("Rolling hills spread noise");
	gettext("2D noise that controls the size/occurrence of rolling hills.");
	gettext("Ridge mountain spread noise");
	gettext("2D noise that controls the size/occurrence of ridged mountain ranges.");
	gettext("Step mountain spread noise");
	gettext("2D noise that controls the size/occurrence of step mountain ranges.");
	gettext("Rolling hill size noise");
	gettext("2D noise that controls the shape/size of rolling hills.");
	gettext("Ridged mountain size noise");
	gettext("2D noise that controls the shape/size of ridged mountains.");
	gettext("Step mountain size noise");
	gettext("2D noise that controls the shape/size of step mountains.");
	gettext("River noise");
	gettext("2D noise that locates the river valleys and channels.");
	gettext("Mountain variation noise");
	gettext("3D noise for mountain overhangs, cliffs, etc. Usually small variations.");
	gettext("Mapgen Flat");
	gettext("Mapgen Flat specific flags");
	gettext("Map generation attributes specific to Mapgen Flat.\nOccasional lakes and hills can be added to the flat world.");
	gettext("Ground level");
	gettext("Y of flat ground.");
	gettext("Lake threshold");
	gettext("Terrain noise threshold for lakes.\nControls proportion of world area covered by lakes.\nAdjust towards 0.0 for a larger proportion.");
	gettext("Lake steepness");
	gettext("Controls steepness/depth of lake depressions.");
	gettext("Hill threshold");
	gettext("Terrain noise threshold for hills.\nControls proportion of world area covered by hills.\nAdjust towards 0.0 for a larger proportion.");
	gettext("Hill steepness");
	gettext("Controls steepness/height of hills.");
	gettext("Terrain noise");
	gettext("Defines location and terrain of optional hills and lakes.");
	gettext("Mapgen Fractal");
	gettext("Mapgen Fractal specific flags");
	gettext("Map generation attributes specific to Mapgen Fractal.\n'terrain' enables the generation of non-fractal terrain:\nocean, islands and underground.");
	gettext("Fractal type");
	gettext("Selects one of 18 fractal types.\n1 = 4D \"Roundy\" Mandelbrot set.\n2 = 4D \"Roundy\" Julia set.\n3 = 4D \"Squarry\" Mandelbrot set.\n4 = 4D \"Squarry\" Julia set.\n5 = 4D \"Mandy Cousin\" Mandelbrot set.\n6 = 4D \"Mandy Cousin\" Julia set.\n7 = 4D \"Variation\" Mandelbrot set.\n8 = 4D \"Variation\" Julia set.\n9 = 3D \"Mandelbrot/Mandelbar\" Mandelbrot set.\n10 = 3D \"Mandelbrot/Mandelbar\" Julia set.\n11 = 3D \"Christmas Tree\" Mandelbrot set.\n12 = 3D \"Christmas Tree\" Julia set.\n13 = 3D \"Mandelbulb\" Mandelbrot set.\n14 = 3D \"Mandelbulb\" Julia set.\n15 = 3D \"Cosine Mandelbulb\" Mandelbrot set.\n16 = 3D \"Cosine Mandelbulb\" Julia set.\n17 = 4D \"Mandelbulb\" Mandelbrot set.\n18 = 4D \"Mandelbulb\" Julia set.");
	gettext("Iterations");
	gettext("Iterations of the recursive function.\nIncreasing this increases the amount of fine detail, but also\nincreases processing load.\nAt iterations = 20 this mapgen has a similar load to mapgen V7.");
	gettext("Scale");
	gettext("(X,Y,Z) scale of fractal in nodes.\nActual fractal size will be 2 to 3 times larger.\nThese numbers can be made very large, the fractal does\nnot have to fit inside the world.\nIncrease these to 'zoom' into the detail of the fractal.\nDefault is for a vertically-squashed shape suitable for\nan island, set all 3 numbers equal for the raw shape.");
	gettext("Offset");
	gettext("(X,Y,Z) offset of fractal from world center in units of 'scale'.\nCan be used to move a desired point to (0, 0) to create a\nsuitable spawn point, or to allow 'zooming in' on a desired\npoint by increasing 'scale'.\nThe default is tuned for a suitable spawn point for Mandelbrot\nsets with default parameters, it may need altering in other\nsituations.\nRange roughly -2 to 2. Multiply by 'scale' for offset in nodes.");
	gettext("Slice w");
	gettext("W coordinate of the generated 3D slice of a 4D fractal.\nDetermines which 3D slice of the 4D shape is generated.\nAlters the shape of the fractal.\nHas no effect on 3D fractals.\nRange roughly -2 to 2.");
	gettext("Julia x");
	gettext("Julia set only.\nX component of hypercomplex constant.\nAlters the shape of the fractal.\nRange roughly -2 to 2.");
	gettext("Julia y");
	gettext("Julia set only.\nY component of hypercomplex constant.\nAlters the shape of the fractal.\nRange roughly -2 to 2.");
	gettext("Julia z");
	gettext("Julia set only.\nZ component of hypercomplex constant.\nAlters the shape of the fractal.\nRange roughly -2 to 2.");
	gettext("Julia w");
	gettext("Julia set only.\nW component of hypercomplex constant.\nAlters the shape of the fractal.\nHas no effect on 3D fractals.\nRange roughly -2 to 2.");
	gettext("Seabed noise");
	gettext("Y-level of seabed.");
	gettext("Mapgen Valleys");
	gettext("Mapgen Valleys specific flags");
	gettext("Map generation attributes specific to Mapgen Valleys.\n'altitude_chill': Reduces heat with altitude.\n'humid_rivers': Increases humidity around rivers.\n'vary_river_depth': If enabled, low humidity and high heat causes rivers\nto become shallower and occasionally dry.\n'altitude_dry': Reduces humidity with altitude.");
	gettext("Altitude chill");
	gettext("The vertical distance over which heat drops by 20 if 'altitude_chill' is\nenabled. Also, the vertical distance over which humidity drops by 10 if\n'altitude_dry' is enabled.");
	gettext("Depth below which you'll find large caves.");
	gettext("Cavern upper limit");
	gettext("Depth below which you'll find giant caverns.");
	gettext("River depth");
	gettext("How deep to make rivers.");
	gettext("River size");
	gettext("How wide to make rivers.");
	gettext("Cave noise #1");
	gettext("Cave noise #2");
	gettext("Filler depth");
	gettext("Terrain height");
	gettext("Base terrain height.");
	gettext("Valley depth");
	gettext("Raises terrain to make valleys around the rivers.");
	gettext("Valley fill");
	gettext("Slope and fill work together to modify the heights.");
	gettext("Valley profile");
	gettext("Amplifies the valleys.");
	gettext("Valley slope");
	gettext("Advanced");
	gettext("Developer Options");
	gettext("Client modding");
	gettext("Enable Lua modding support on client.\nThis support is experimental and API can change.");
	gettext("Main menu script");
	gettext("Replaces the default main menu with a custom one.");
	gettext("Mod Security");
	gettext("Enable mod security");
	gettext("Prevent mods from doing insecure things like running shell commands.");
	gettext("Trusted mods");
	gettext("Comma-separated list of trusted mods that are allowed to access insecure\nfunctions even when mod security is on (via request_insecure_environment()).");
	gettext("HTTP mods");
	gettext("Comma-separated list of mods that are allowed to access HTTP APIs, which\nallow them to upload and download data to/from the internet.");
	gettext("Debugging");
	gettext("Debug log level");
	gettext("Level of logging to be written to debug.txt:\n-    <nothing> (no logging)\n-    none (messages with no level)\n-    error\n-    warning\n-    action\n-    info\n-    verbose\n-    trace");
	gettext("Debug log file size threshold");
	gettext("If the file size of debug.txt exceeds the number of megabytes specified in\nthis setting when it is opened, the file is moved to debug.txt.1,\ndeleting an older debug.txt.1 if it exists.\ndebug.txt is only moved if this setting is positive.");
	gettext("Chat log level");
	gettext("Minimal level of logging to be written to chat.");
	gettext("Deprecated Lua API handling");
	gettext("Handling for deprecated Lua API calls:\n-    none: Do not log deprecated calls\n-    log: mimic and log backtrace of deprecated call (default).\n-    error: abort on usage of deprecated call (suggested for mod developers).");
	gettext("Random input");
	gettext("Enable random user input (only used for testing).");
	gettext("Random mod load order");
	gettext("Enable random mod loading (mainly used for testing).");
	gettext("Mod channels");
	gettext("Enable mod channels support.");
	gettext("Mod Profiler");
	gettext("Load the game profiler");
	gettext("Load the game profiler to collect game profiling data.\nProvides a /profiler command to access the compiled profile.\nUseful for mod developers and server operators.");
	gettext("Measure time samples");
	gettext("Disable if you only want to instrument with Tracy.");
	gettext("Capture Tracy zones");
	gettext("Capture Tracy zones with the profiler's instrumentation.\nYou need to build with Tracy to use this, see doc/developing/profiling.md\nfor details.");
	gettext("Default report format");
	gettext("The default format in which profiles are being saved,\nwhen calling `/profiler save [format]` without format.");
	gettext("Report path");
	gettext("The file path relative to your world path in which profiles will be saved to.");
	gettext("Entity methods");
	gettext("Instrument the methods of entities on registration.");
	gettext("Active Block Modifiers");
	gettext("Instrument the action function of Active Block Modifiers on registration.");
	gettext("Loading Block Modifiers");
	gettext("Instrument the action function of Loading Block Modifiers on registration.");
	gettext("Chat commands");
	gettext("Instrument chat commands on registration.");
	gettext("Global callbacks");
	gettext("Instrument global callback functions on registration.\n(anything you pass to a core.register_*() function)");
	gettext("Builtin");
	gettext("Instrument builtin.\nThis is usually only needed by core/builtin contributors");
	gettext("Profiler");
	gettext("Have the profiler instrument itself:\n* Instrument an empty function.\nThis estimates the overhead, that instrumentation is adding (+1 function call).\n* Instrument the sampler being used to update the statistics.");
	gettext("Engine Profiler");
	gettext("Engine profiling data print interval");
	gettext("Print the engine's profiling data in regular intervals (in seconds).\n0 = disable. Useful for developers.");
	gettext("OpenGL debug");
	gettext("Enables debug and error-checking in the OpenGL driver.");
	gettext("Shader path");
	gettext("Path to shader directory. If no path is defined, default location will be used.");
	gettext("Video driver");
	gettext("The rendering back-end.\nNote: A restart is required after changing this!\nOpenGL is the default for desktop, and OGLES2 for Android.");
	gettext("Transparency Sorting Distance");
	gettext("Distance in nodes at which transparency depth sorting is enabled.\nUse this to limit the performance impact of transparency depth sorting.\nSet to 0 to disable it entirely.");
	gettext("Transparency Sorting Group by Buffers");
	gettext("Draw transparency sorted triangles grouped by their mesh buffers.\nThis breaks transparency sorting between mesh buffers, but avoids situations\nwhere transparency sorting would be very slow otherwise.");
	gettext("Cloud radius");
	gettext("Radius of cloud area stated in number of 64 node cloud squares.\nValues larger than 26 will start to produce sharp cutoffs at cloud area corners.");
	gettext("Mapblock mesh generation delay");
	gettext("Delay between mesh updates on the client in ms. Increasing this will slow\ndown the rate of mesh updates, which can help reduce jitter.");
	gettext("Mapblock mesh generation threads");
	gettext("Number of threads to use for mesh generation.\nValue of 0 (default) will let Luanti automatically choose the number of threads.");
	gettext("Minimum vertex count for mesh buffers");
	gettext("All mesh buffers with less than this number of vertices will be merged\nduring map rendering. This improves rendering performance.");
	gettext("Minimap scan height");
	gettext("True = 256\nFalse = 128\nUsable to make minimap smoother on slower machines.");
	gettext("World-aligned textures mode");
	gettext("Textures on a node may be aligned either to the node or to the world.\nThe former mode suits better things like machines, furniture, etc., while\nthe latter makes stairs and microblocks fit surroundings better.\nHowever, as this possibility is new, thus may not be used by older servers,\nthis option allows enforcing it for certain node types. Note though that\nthat is considered EXPERIMENTAL and may not work properly.");
	gettext("Autoscaling mode");
	gettext("World-aligned textures may be scaled to span several nodes. However,\nthe server may not send the scale you want, especially if you use\na specially-designed texture pack; with this option, the client tries\nto determine the scale automatically based on the texture size.\nSee also texture_min_size.\nWarning: This option is EXPERIMENTAL!");
	gettext("Base texture size");
	gettext("When using bilinear/trilinear filtering, low-resolution textures\ncan be blurred, so this option automatically upscales them to preserve\ncrisp pixels. This defines the minimum texture size for the upscaled textures;\nhigher values look sharper, but require more memory.\nThis setting is ONLY applied if any of the mentioned filters are enabled.\nThis is also used as the base node texture size for world-aligned\ntexture autoscaling.");
	gettext("Client Mesh Chunksize");
	gettext("Side length of a cube of map blocks that the client will consider together\nwhen generating meshes.\nLarger values increase the utilization of the GPU by reducing the number of\ndraw calls, benefiting especially high-end GPUs.\nSystems with a low-end GPU (or no GPU) would benefit from smaller values.");
	gettext("Color depth for post-processing texture");
	gettext("Decide the color depth of the texture used for the post-processing pipeline.\nReducing this can improve performance, but some effects (e.g. debanding)\nrequire more than 8 bits to work.");
	gettext("Poisson filtering");
	gettext("Enable Poisson disk filtering.\nOn true uses Poisson disk to make \"soft shadows\". Otherwise uses PCF filtering.");
	gettext("Map shadows update frames");
	gettext("Spread a complete update of the shadow map over a given number of frames.\nHigher values might make shadows laggy, lower values\nwill consume more resources.");
	gettext("Enable Bloom Debug");
	gettext("Set to true to render debugging breakdown of the bloom effect.\nIn debug mode, the screen is split into 4 quadrants:\ntop-left - processed base image, top-right - final image\nbottom-left - raw base image, bottom-right - bloom texture.");
	gettext("Sound");
	gettext("Sound Extensions Blacklist");
	gettext("Comma-separated list of AL and ALC extensions that should not be used.\nUseful for testing. See al_extensions.[h,cpp] for details.");
	gettext("Font");
	gettext("Font bold by default");
	gettext("Font italic by default");
	gettext("Font shadow");
	gettext("Shadow offset (in pixels) of the default font. If 0, then shadow will not be drawn.");
	gettext("Font shadow alpha");
	gettext("Opaqueness (alpha) of the shadow behind the default font, between 0 and 255.");
	gettext("Font size");
	gettext("Font size of the default font where 1 unit = 1 pixel at 96 DPI");
	gettext("Font size divisible by");
	gettext("For pixel-style fonts that do not scale well, this ensures that font sizes used\nwith this font will always be divisible by this value, in pixels. For instance,\na pixel font 16 pixels tall should have this set to 16, so it will only ever be\nsized 16, 32, 48, etc., so a mod requesting a size of 25 will get 32.");
	gettext("Regular font path");
	gettext("Path to the default font. Must be a TrueType font.\nThe fallback font will be used if the font cannot be loaded.");
	gettext("Bold font path");
	gettext("Italic font path");
	gettext("Bold and italic font path");
	gettext("Monospace font size");
	gettext("Font size of the monospace font where 1 unit = 1 pixel at 96 DPI");
	gettext("Monospace font size divisible by");
	gettext("Monospace font path");
	gettext("Path to the monospace font. Must be a TrueType font.\nThis font is used for e.g. the console and profiler screen.");
	gettext("Bold monospace font path");
	gettext("Italic monospace font path");
	gettext("Bold and italic monospace font path");
	gettext("Fallback font path");
	gettext("Path of the fallback font. Must be a TrueType font.\nThis font will be used for certain languages or if the default font is unavailable.");
	gettext("Lighting");
	gettext("Light curve low gradient");
	gettext("Gradient of light curve at minimum light level.\nControls the contrast of the lowest light levels.");
	gettext("Light curve high gradient");
	gettext("Gradient of light curve at maximum light level.\nControls the contrast of the highest light levels.");
	gettext("Light curve boost");
	gettext("Strength of light curve boost.\nThe 3 'boost' parameters define a range of the light\ncurve that is boosted in brightness.");
	gettext("Light curve boost center");
	gettext("Center of light curve boost range.\nWhere 0.0 is minimum light level, 1.0 is maximum light level.");
	gettext("Light curve boost spread");
	gettext("Spread of light curve boost range.\nControls the width of the range to be boosted.\nStandard deviation of the light curve boost Gaussian.");
	gettext("IPv6");
	gettext("Enable IPv6 support (for both client and server).\nRequired for IPv6 connections to work at all.");
	gettext("Prometheus listener address");
	gettext("Prometheus listener address.\nIf Luanti is compiled with Prometheus support, this setting\nenables the metrics listener for Prometheus on that address.\nBy default you can fetch metrics from http://127.0.0.1:30000/metrics.\nAn empty value disables the metrics listener.");
	gettext("Maximum size of the client's outgoing chat queue");
	gettext("Maximum size of the client's outgoing chat queue.\n0 to disable queueing and -1 to make the queue size unlimited.");
	gettext("Mapblock unload timeout");
	gettext("Timeout for client to remove unused map data from memory, in seconds.");
	gettext("Mapblock limit");
	gettext("Maximum number of mapblocks for client to be kept in memory.\nNote that there is an internal dynamic minimum number of blocks that\nwon't be deleted, depending on the current view range.\nSet to -1 for no limit.");
	gettext("Maximum simultaneous block sends per client");
	gettext("Maximum number of blocks that are simultaneously sent per client.\nThe maximum total count is calculated dynamically:\nmax_total = ceil((#clients + max_users) * per_client / 4)");
	gettext("Delay in sending blocks after building");
	gettext("To reduce lag, block transfers are slowed down when a player is building something.\nThis determines how long they are slowed down after placing or removing a node.");
	gettext("Max. packets per iteration");
	gettext("Maximum number of packets sent per send step in the low-level networking code.\nYou generally don't need to change this, however busy servers may benefit from a higher number.");
	gettext("Map Compression Level for Network Transfer");
	gettext("Compression level to use when sending mapblocks to the client.\n-1 - use default compression level\n0 - least compression, fastest\n9 - best compression, slowest");
	gettext("Chat message format");
	gettext("Format of player chat messages. The following strings are valid placeholders:\n@name, @message, @timestamp (optional)");
	gettext("Chat command time message threshold");
	gettext("If the execution of a chat command takes longer than this specified time in\nseconds, add the time information to the chat command message");
	gettext("Shutdown message");
	gettext("A message to be displayed to all clients when the server shuts down.");
	gettext("Crash message");
	gettext("A message to be displayed to all clients when the server crashes.");
	gettext("Ask to reconnect after crash");
	gettext("Whether to ask clients to reconnect after a (Lua) crash.\nSet this to true if your server is set up to restart automatically.");
	gettext("Server/Env Performance");
	gettext("Dedicated server step");
	gettext("Length of a server tick (the interval at which everything is generally updated),\nstated in seconds.\nDoes not apply to sessions hosted from the client menu.\nThis is a lower bound, i.e. server steps may not be shorter than this, but\nthey are often longer.");
	gettext("Unlimited player transfer distance");
	gettext("Whether players are shown to clients without any range limit.\nDeprecated, use the setting player_transfer_distance instead.");
	gettext("Player transfer distance");
	gettext("Defines the maximal player transfer distance in blocks (0 = unlimited).");
	gettext("Active object send range");
	gettext("From how far clients know about objects, stated in mapblocks (16 nodes).\n\nSetting this larger than active_block_range will also cause the server\nto maintain active objects up to this distance in the direction the\nplayer is looking. (This can avoid mobs suddenly disappearing from view)");
	gettext("Active block range");
	gettext("The radius of the volume of blocks around every player that is subject to the\nactive block stuff, stated in mapblocks (16 nodes).\nIn active blocks objects are loaded and ABMs run.\nThis is also the minimum range in which active objects (mobs) are maintained.\nThis should be configured together with active_object_send_range_blocks.");
	gettext("Max block send distance");
	gettext("From how far blocks are sent to clients, stated in mapblocks (16 nodes).");
	gettext("Maximum forceloaded blocks");
	gettext("Default maximum number of forceloaded mapblocks.\nSet this to -1 to disable the limit.");
	gettext("Map save interval");
	gettext("Interval of saving important changes in the world, stated in seconds.");
	gettext("Unload unused server data");
	gettext("How long the server will wait before unloading unused mapblocks, stated in seconds.\nHigher value is smoother, but will use more RAM.");
	gettext("Maximum objects per block");
	gettext("Maximum number of statically stored objects in a block.");
	gettext("Active block management interval");
	gettext("Length of time between active block management cycles, stated in seconds.");
	gettext("ABM interval");
	gettext("Length of time between Active Block Modifier (ABM) execution cycles, stated in seconds.");
	gettext("ABM time budget");
	gettext("The time budget allowed for ABMs to execute on each step\n(as a fraction of the ABM Interval)");
	gettext("NodeTimer interval");
	gettext("Length of time between NodeTimer execution cycles, stated in seconds.");
	gettext("Liquid loop max");
	gettext("Max liquids processed per step.");
	gettext("Liquid queue purge time");
	gettext("The time (in seconds) that the liquids queue may grow beyond processing\ncapacity until an attempt is made to decrease its size by dumping old queue\nitems.  A value of 0 disables the functionality.");
	gettext("Liquid update tick");
	gettext("Liquid update interval in seconds.");
	gettext("Block send optimize distance");
	gettext("At this distance the server will aggressively optimize which blocks are sent to\nclients.\nSmall values potentially improve performance a lot, at the expense of visible\nrendering glitches (some blocks might not be rendered correctly in caves).\nSetting this to a value greater than max_block_send_distance disables this\noptimization.\nStated in MapBlocks (16 nodes).");
	gettext("Server-side occlusion culling");
	/* xgettext:no-c-format */ gettext("If enabled, the server will perform map block occlusion culling based on\non the eye position of the player. This can reduce the number of blocks\nsent to the client by 50-80%. Clients will no longer receive most\ninvisible blocks, so that the utility of noclip mode is reduced.");
	gettext("Block cull optimize distance");
	gettext("At this distance the server will perform a simpler and cheaper occlusion check.\nSmaller values potentially improve performance, at the expense of temporarily visible\nrendering glitches (missing blocks).\nThis is especially useful for very large viewing range (upwards of 500).\nStated in MapBlocks (16 nodes).");
	gettext("Chunk size");
	gettext("Size of mapchunks generated by mapgen, stated in mapblocks (16 nodes).\nWARNING: There is no benefit, and there are several dangers, in\nincreasing this value above 5.\nReducing this value increases cave and dungeon density.\nAltering this value is for special usage, leaving it unchanged is\nrecommended.");
	gettext("Mapgen debug");
	gettext("Dump the mapgen debug information.");
	gettext("Absolute limit of queued blocks to emerge");
	gettext("Maximum number of blocks that can be queued for loading.");
	gettext("Per-player limit of queued blocks load from disk");
	gettext("Maximum number of blocks to be queued that are to be loaded from file.\nThis limit is enforced per player.");
	gettext("Per-player limit of queued blocks to generate");
	gettext("Maximum number of blocks to be queued that are to be generated.\nThis limit is enforced per player.");
	gettext("Number of emerge threads");
	gettext("Number of emerge threads to use.\nValue 0:\n-    Automatic selection. The number of emerge threads will be\n-    'number of processors - 2', with a lower limit of 1.\nAny other value:\n-    Specifies the number of emerge threads, with a lower limit of 1.\nWARNING: Increasing the number of emerge threads increases engine mapgen\nspeed, but this may harm game performance by interfering with other\nprocesses, especially in singleplayer and/or when running Lua code in\n'on_generated'. For many users the optimum setting may be '1'.");
	gettext("cURL");
	gettext("cURL interactive timeout");
	gettext("Maximum time an interactive request (e.g. server list fetch) may take, stated in milliseconds.");
	gettext("cURL parallel limit");
	gettext("Limits number of parallel HTTP requests. Affects:\n-    Media fetch if server uses remote_media setting.\n-    Serverlist download and server announcement.\n-    Downloads performed by main menu (e.g. mod manager).\nOnly has an effect if compiled with cURL.");
	gettext("cURL file download timeout");
	gettext("Maximum time a file download (e.g. a mod download) may take, stated in milliseconds.");
	gettext("Client Debugging");
	gettext("Toggle camera update");
	gettext("Key for toggling the camera update. Only usable with 'debug' privilege.");
	gettext("Quicktune: select previous entry");
	gettext("Key for switching to the previous entry in Quicktune.");
	gettext("Quicktune: select next entry");
	gettext("Key for switching to the next entry in Quicktune.");
	gettext("Quicktune: decrement value");
	gettext("Key for decrementing the selected value in Quicktune.");
	gettext("Quicktune: increment value");
	gettext("Key for incrementing the selected value in Quicktune.");
	gettext("Miscellaneous");
	gettext("Chat weblinks");
	gettext("Clickable weblinks (middle-click or Ctrl+left-click) enabled in chat console output.");
	gettext("Ignore world errors");
	gettext("If enabled, invalid world data won't cause the server to shut down.\nOnly enable this if you know what you are doing.");
	gettext("Display Density Scaling Factor");
	gettext("Adjust the detected display density, used for scaling UI elements.");
	gettext("Enable console window");
	gettext("Windows systems only: Start Luanti with the command line window in the background.\nContains the same information as the file debug.txt (default name).");
	gettext("Max. clearobjects extra blocks");
	gettext("Number of extra blocks that can be loaded by /clearobjects at once.\nThis is a trade-off between SQLite transaction overhead and\nmemory consumption (4096=100MB, as a rule of thumb).");
	gettext("Map directory");
	gettext("World directory (everything in the world is stored here).\nNot needed if starting from the main menu.");
	gettext("Synchronous SQLite");
	gettext("See https://www.sqlite.org/pragma.html#pragma_synchronous");
	gettext("Map Compression Level for Disk Storage");
	gettext("Compression level to use when saving mapblocks to disk.\n-1 - use default compression level\n0 - least compression, fastest\n9 - best compression, slowest");
	gettext("Connect to external media server");
	gettext("Enable usage of remote media server (if provided by server).\nRemote servers offer a significantly faster way to download media (e.g. textures)\nwhen connecting to the server.");
	gettext("Serverlist file");
	gettext("File in client/serverlist/ that contains your favorite servers displayed in the\nMultiplayer Tab.");
	gettext("Gamepads");
	gettext("Enable joysticks");
	gettext("Enable joysticks. Requires a restart to take effect");
	gettext("Joystick ID");
	gettext("The identifier of the joystick to use");
	gettext("Joystick type");
	gettext("The type of joystick");
	gettext("Joystick button repetition interval");
	gettext("The time in seconds it takes between repeated events\nwhen holding down a joystick button combination.");
	gettext("Joystick dead zone");
	gettext("The dead zone of the joystick");
	gettext("Joystick frustum sensitivity");
	gettext("The sensitivity of the joystick axes for moving the\nin-game view frustum around.");
}
