dofile(core.get_modpath("testentities").."/visuals.lua")
dofile(core.get_modpath("testentities").."/observers.lua")
dofile(core.get_modpath("testentities").."/selectionbox.lua")
dofile(core.get_modpath("testentities").."/armor.lua")
dofile(core.get_modpath("testentities").."/pointable.lua")
